@@ -872,7 +872,7 @@ comsub(Expand *xp, const char *cp)
 		ofd1 = savefd(1, 0);	/* fd 1 may be closed... */
 		ksh_dup2(pv[1], 1, FALSE);
 		close(pv[1]);
-		execute(t, XFORK|XXCOM|XPIPEO);
+		execute(t, XFORK | XXCOM | XPIPEO, NULL);
 		restfd(1, ofd1);
 		startlast();
 		xp->split = 1;	/* waitlast() */
