@@ -15,7 +15,8 @@
 #endif /* KSH */
 
 static int	comexec	 ARGS((struct op *t, struct tbl *volatile tp, char **ap,
-			      int volatile flags));
+			      int volatile flags, volatile int *xerrok));
+
 static void	scriptexec ARGS((struct op *tp, char **ap));
 static int	call_builtin ARGS((struct tbl *tp, char **wp));
 static int	iosetup ARGS((struct ioword *iop, struct tbl *tp));
@@ -61,13 +62,10 @@ fd_clexec(fd)
  * execute command tree
  */
 int
-execute(t, flags)
-	struct op * volatile t;
-	volatile int flags;	/* if XEXEC don't fork */
+execute(struct op * volatile t, volatile int flags, volatile int * volatile xerrok)
 {
 	int i;
-	volatile int rv = 0;
-	volatile int rv_prop = 0; /* rv being propogated or newly generated? */
+	volatile int rv = 0, dummy = 0;
 	int pv[2];
 	char ** volatile ap;
 	char *s, *cp;
@@ -75,7 +73,11 @@ execute(t, flags)
 	struct tbl *tp = NULL;
 
 	if (t == NULL)
-		return 0;
+		return (0);
+
+	/* Caller doesn't care if XERROK should propagate. */
+	if (xerrok == NULL)
+		xerrok = &dummy;
 
 	/* Is this the end of a pipeline?  If so, we want to evaluate the
 	 * command arguments
@@ -86,16 +88,12 @@ execute(t, flags)
 	}
 	 */
 	if ((flags&XFORK) && !(flags&XEXEC) && t->type != TPIPE)
-#ifdef SILLY_FEATURES
-		return exchild(t, flags & ~XTIME, -1); /* run in sub-process */
-#else /* SILLY_FEATURES */
-		return exchild(t, flags, -1); /* run in sub-process */
-#endif /* SILLY_FEATURES */
+		return exchild(t, flags, xerrok, -1); /* run in sub-process */
 
 	newenv(E_EXEC);
 	if (trap)
 		runtraps(0);
- 
+
 	if (t->type == TCOM) {
 		/* Clear subst_exstat before argument expansion.  Used by
 		 * null commands (see comexec() and c_eval()) and by c_set().
@@ -108,11 +106,6 @@ execute(t, flags)
 		 * and assignments last..
 		 */
 		ap = eval((const char **)t->args, t->evalflags | DOBLANK | DOGLOB | DOTILDE);
-#ifdef SILLY_FEATURES
-		if (flags & XTIME)
-			/* Allow option parsing (bizarre, but POSIX) */
-			timex_hook(t, &ap);
-#endif /* SILLY_FEATURES */
 		if (Flag(FXTRACE) && ap[0]) {
 			shf_fprintf(shl_out, "%s",
 				PS4_SUBSTITUTE(str_val(global("PS4"))));
@@ -152,12 +145,11 @@ execute(t, flags)
 	
 	switch(t->type) {
 	  case TCOM:
-		rv = comexec(t, tp, ap, flags);
+		rv = comexec(t, tp, ap, flags, xerrok);
 		break;
 
 	  case TPAREN:
-		rv = execute(t->left, flags|XFORK);
-		rv_prop = 1;
+		rv = execute(t->left, flags | XFORK, xerrok);
 		break;
 
 	  case TPIPE:
@@ -174,7 +166,8 @@ execute(t, flags)
 			 *    (: ; cat /etc/termcap) | sleep 1
 			 *  will hang forever).
 			 */
-			exchild(t->left, flags|XPIPEO|XCCLOSE, pv[0]);
+			exchild(t->left, flags | XPIPEO | XCCLOSE,
+			    NULL, pv[0]);
 			(void) ksh_dup2(pv[0], 0, FALSE); /* stdin of next */
 			closepipe(pv);
 			flags |= XPIPEI;
@@ -183,99 +176,43 @@ execute(t, flags)
 		restfd(1, e->savefd[1]); /* stdout of last */
 		e->savefd[1] = 0; /* no need to re-restore this */
 		/* Let exchild() close 0 in parent, after fork, before wait */
-		i = exchild(t, flags|XPCLOSE, 0);
+		i = exchild(t, flags | XPCLOSE, xerrok, 0);
 		if (!(flags&XBGND) && !(flags&XXCOM))
 			rv = i;
 		break;
 
-	  case TLIST:
+	case TLIST:
 		while (t->type == TLIST) {
-			execute(t->left, flags & XERROK);
+			execute(t->left, flags & XERROK, NULL);
 			t = t->right;
 		}
-		rv = execute(t, flags & XERROK);
+		rv = execute(t, flags & XERROK, xerrok);
 		break;
 
-#ifdef KSH
-	  case TCOPROC:
-	  {
-		sigset_t	omask;
-
-		/* Block sigchild as we are using things changed in the
-		 * signal handler
-		 */
-		sigprocmask(SIG_BLOCK, &sm_sigchld, &omask);
-		e->type = E_ERRH;
-		i = sigsetjmp(e->jbuf, 0);
-		if (i) {
-			sigprocmask(SIG_SETMASK, &omask, (sigset_t *) 0);
-			quitenv();
-			unwind(i);
-			/*NOTREACHED*/
-		}
-		/* Already have a (live) co-process? */
-		if (coproc.job && coproc.write >= 0)
-			errorf("coprocess already exists");
-
-		/* Can we re-use the existing co-process pipe? */
-		coproc_cleanup(TRUE);
-
-		/* do this before opening pipes, in case these fail */
-		e->savefd[0] = savefd(0, 0);
-		e->savefd[1] = savefd(1, 0);
-
-		openpipe(pv);
-		ksh_dup2(pv[0], 0, FALSE);
-		close(pv[0]);
-		coproc.write = pv[1];
-		coproc.job = (void *) 0;
-
-		if (coproc.readw >= 0)
-			ksh_dup2(coproc.readw, 1, FALSE);
-		else {
-			openpipe(pv);
-			coproc.read = pv[0];
-			ksh_dup2(pv[1], 1, FALSE);
-			coproc.readw = pv[1];	 /* closed before first read */
-			coproc.njobs = 0;
-			/* create new coprocess id */
-			++coproc.id;
-		}
-		sigprocmask(SIG_SETMASK, &omask, (sigset_t *) 0);
-		e->type = E_EXEC; /* no more need for error handler */
-
-		/* exchild() closes coproc.* in child after fork,
-		 * will also increment coproc.njobs when the
-		 * job is actually created.
-		 */
-		flags &= ~XEXEC;
-		exchild(t->left, flags|XBGND|XFORK|XCOPROC|XCCLOSE,
-			coproc.readw);
-		break;
-	  }
-#endif /* KSH */
 
 	  case TASYNC:
 		/* XXX non-optimal, I think - "(foo &)", forks for (),
 		 * forks again for async...  parent should optimize
 		 * this to "foo &"...
 		 */
-		rv = execute(t->left, (flags&~XEXEC)|XBGND|XFORK);
+		rv = execute(t->left, (flags&~XEXEC)|XBGND|XFORK, xerrok);
 		break;
 
-	  case TOR:
-	  case TAND:
-		rv = execute(t->left, XERROK);
-		if (t->right != NULL && (rv == 0) == (t->type == TAND))
-			rv = execute(t->right, flags & XERROK);
-		else
-			flags |= XERROK;
-		rv_prop = 1;
-		break;
-
-	  case TBANG:
-		rv = !execute(t->right, XERROK);
+	case TOR:
+	case TAND:
+		rv = execute(t->left, XERROK, xerrok);
+		if ((rv == 0) == (t->type == TAND))
+			rv = execute(t->right, XERROK, xerrok);
 		flags |= XERROK;
+		if (xerrok)
+			*xerrok = 1;
+		break;
+
+	case TBANG:
+		rv = !execute(t->right, XERROK, xerrok);
+		flags |= XERROK;
+		if (xerrok)
+			*xerrok = 1;
 		break;
 
 #ifdef KSH
@@ -295,7 +232,7 @@ execute(t, flags)
 	    }
 #endif /* KSH */
 
-	  case TFOR:
+	case TFOR:
 #ifdef KSH
 	  case TSELECT:
 	    {
@@ -309,9 +246,8 @@ execute(t, flags)
 			i = sigsetjmp(e->jbuf, 0);
 			if (!i)
 				break;
-			if ((e->flags&EF_BRKCONT_PASS)
-			    || (i != LBREAK && i != LCONTIN))
-			{
+			if ((e->flags&EF_BRKCONT_PASS) ||
+			    (i != LBREAK && i != LCONTIN)) {
 				quitenv();
 				unwind(i);
 			} else if (i == LBREAK) {
@@ -320,11 +256,10 @@ execute(t, flags)
 			}
 		}
 		rv = 0; /* in case of a continue */
-		rv_prop = 1;
 		if (t->type == TFOR) {
 			while (*ap != NULL) {
 				setstr(global(t->str), *ap++, KSH_UNWIND_ERROR);
-				rv = execute(t->left, flags & XERROK);
+				rv = execute(t->left, flags & XERROK, xerrok);
 			}
 		}
 #ifdef KSH
@@ -332,28 +267,26 @@ execute(t, flags)
 			for (;;) {
 				if (!(cp = do_selectargs(ap, is_first))) {
 					rv = 1;
-					rv_prop = 0;
 					break;
 				}
 				is_first = FALSE;
 				setstr(global(t->str), cp, KSH_UNWIND_ERROR);
-				rv = execute(t->left, flags & XERROK);
+				rv = execute(t->left, flags & XERROK, xerrok);
 			}
 		}
 	    }
 #endif /* KSH */
 		break;
 
-	  case TWHILE:
-	  case TUNTIL:
+	case TWHILE:
+	case TUNTIL:
 		e->type = E_LOOP;
 		while (1) {
 			i = sigsetjmp(e->jbuf, 0);
 			if (!i)
 				break;
-			if ((e->flags&EF_BRKCONT_PASS)
-			    || (i != LBREAK && i != LCONTIN))
-			{
+			if ((e->flags&EF_BRKCONT_PASS) ||
+			    (i != LBREAK && i != LCONTIN)) {
 				quitenv();
 				unwind(i);
 			} else if (i == LBREAK) {
@@ -362,63 +295,44 @@ execute(t, flags)
 			}
 		}
 		rv = 0; /* in case of a continue */
-		while ((execute(t->left, XERROK) == 0) == (t->type == TWHILE))
-			rv = execute(t->right, flags & XERROK);
-		rv_prop = 1;
+		while ((execute(t->left, XERROK, NULL) == 0) ==
+		    (t->type == TWHILE))
+			rv = execute(t->right, flags & XERROK, xerrok);
 		break;
 
-	  case TIF:
-	  case TELIF:
+	case TIF:
+	case TELIF:
 		if (t->right == NULL)
 			break;	/* should be error */
-		rv = execute(t->left, XERROK) == 0 ?
-			execute(t->right->left, flags & XERROK) :
-			execute(t->right->right, flags & XERROK);
-		rv_prop = 1;
+		rv = execute(t->left, XERROK, NULL) == 0 ?
+		    execute(t->right->left, flags & XERROK, xerrok) :
+		    execute(t->right->right, flags & XERROK, xerrok);
 		break;
 
-	  case TCASE:
+	case TCASE:
 		cp = evalstr(t->str, DOTILDE);
 		for (t = t->left; t != NULL && t->type == TPAT; t = t->right)
 		    for (ap = t->vars; *ap; ap++)
-			if ((s = evalstr(*ap, DOTILDE|DOPAT))
-			    && gmatchx(cp, s, FALSE))
+			if ((s = evalstr(*ap, DOTILDE|DOPAT)) &&
+			    gmatchx(cp, s, FALSE))
 				goto Found;
 		break;
-	  Found:
-		rv = execute(t->left, flags & XERROK);
-		rv_prop = 1;
+ Found:
+		rv = execute(t->left, flags & XERROK, xerrok);
 		break;
 
-	  case TBRACE:
-		rv = execute(t->left, flags & XERROK);
-		rv_prop = 1;
+	case TBRACE:
+		rv = execute(t->left, flags & XERROK, xerrok);
 		break;
 
-	  case TFUNCT:
+	case TFUNCT:
 		rv = define(t->str, t);
 		break;
 
-#ifdef SILLY_FEATURES
-	  case TTIME:
-		/* Clear XEXEC so nested execute() call doesn't exit
-		 * (allows "ls -l | time grep foo").
-		 */
-		rv = timex(t, flags & ~XEXEC);
-		rv_prop = 1;
-		break;
-#endif /* SILLY_FEATURES */
 
-	  case TEXEC:		/* an eval'd TCOM */
+	case TEXEC:		/* an eval'd TCOM */
 		s = t->args[0];
 		ap = makenv();
-#ifndef F_SETFD
-		for (i = 0; i < sizeof(clexec_tab); i++)
-			if (clexec_tab[i]) {
-				close(i);
-				clexec_tab[i] = 0;
-			}
-#endif
 		restoresigs();
 		cleanup_proc_env();
 		/* XINTACT bit is for OS2 */
@@ -428,17 +342,18 @@ execute(t, flags)
 		else
 			errorf("%s: %s", s, strerror(errno));
 	}
-    Break:
+ Break:
 	exstat = rv;
 
 	quitenv();		/* restores IO */
 	if ((flags&XEXEC))
 		unwind(LEXIT);	/* exit child */
-	if (rv != 0 && !rv_prop && !(flags & XERROK)) {
+	if (rv != 0 && !(flags & XERROK) &&
+	    (xerrok == NULL || !*xerrok)) {
 		if (Flag(FERREXIT))
 			unwind(LERROR);
 	}
-	return rv;
+	return (rv);
 }
 
 /*
@@ -446,7 +361,7 @@ execute(t, flags)
  */
 
 static int
-comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
+comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags, volatile int *xerrok)
 {
 	int i;
 	int rv = 0;
@@ -462,11 +377,11 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 	 * before we can do a path search (in case the assignments change
 	 * PATH).
 	 * Odd cases:
-	 *   FOO=bar exec > /dev/null		FOO is kept but not exported
-	 *   FOO=bar exec foobar		FOO is exported
-	 *   FOO=bar command exec > /dev/null	FOO is neither kept nor exported
-	 *   FOO=bar command			FOO is neither kept nor exported
-	 *   PATH=... foobar			use new PATH in foobar search
+	 *	FOO=bar exec >/dev/null		FOO is kept but not exported
+	 *	FOO=bar exec foobar		FOO is exported
+	 *	FOO=bar command exec >/dev/null	FOO is neither kept nor exported
+	 *	FOO=bar command			FOO is neither kept nor exported
+	 *	PATH=... foobar			use new PATH in foobar search
 	 */
 	keepasn_ok = 1;
 	while (tp && tp->type == CSHELL) {
@@ -562,7 +477,7 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 	}
 
 	switch (tp->type) {
-	  case CSHELL:			/* shell built-in */
+	case CSHELL:			/* shell built-in */
 		rv = call_builtin(tp, ap);
 		break;
 
@@ -588,7 +503,7 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 				}
 				break;
 			}
-			if (include(tp->u.fpath, 0, (char **) 0, 0) < 0) {
+			if (include(tp->u.fpath, 0, NULL, 0) < 0) {
 				warningf(TRUE,
 			    "%s: can't open function definition file %s - %s",
 					cp, tp->u.fpath, strerror(errno));
@@ -629,15 +544,15 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 		e->type = E_FUNC;
 		i = sigsetjmp(e->jbuf, 0);
 		if (i == 0) {
-			/* seems odd to pass XERROK here, but at&t ksh does */
-			exstat = execute(tp->val.t, flags & XERROK);
+			/* seems odd to pass XERROK here, but AT&T ksh does */
+			exstat = execute(tp->val.t, flags & XERROK, xerrok);
 			i = LRETURN;
 		}
 		poshname = old_kshname;
 /*		Flag(FXTRACE) = old_xflag; */
 		tp->flag = (tp->flag & ~FINUSE) | old_inuse;
-		/* Were we deleted while executing?  If so, free the execution
-		 * tree.  todo: Unfortunately, the table entry is never re-used
+		/* Were we deleted while executing? If so, free the execution
+		 * tree. todo: Unfortunately, the table entry is never re-used
 		 * until the lookup table is expanded.
 		 */
 		if ((tp->flag & (FDELETE|FINUSE)) == FDELETE) {
@@ -648,30 +563,30 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 			tp->flag = 0;
 		}
 		switch (i) {
-		  case LRETURN:
-		  case LERROR:
+		case LRETURN:
+		case LERROR:
 			rv = exstat;
 			break;
-		  case LINTR:
-		  case LEXIT:
-		  case LLEAVE:
-		  case LSHELL:
+		case LINTR:
+		case LEXIT:
+		case LLEAVE:
+		case LSHELL:
 			quitenv();
 			unwind(i);
 			/*NOTREACHED*/
-		  default:
+		default:
 			quitenv();
 			internal_errorf(1, "CFUNC %d", i);
 		}
 		break;
-	  }
+	}
 
-	  case CEXEC:		/* executable command */
-	  case CTALIAS:		/* tracked alias */
+	case CEXEC:		/* executable command */
+	case CTALIAS:		/* tracked alias */
 		if (!(tp->flag&ISSET)) {
 			/* errno_ will be set if the named command was found
 			 * but could not be executed (permissions, no execute
-			 * bit, directory, etc).  Print out a (hopefully)
+			 * bit, directory, etc). Print out a (hopefully)
 			 * useful error message and set the exit status to 126.
 			 */
 			if (tp->u2.errno_) {
@@ -688,8 +603,8 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 #ifdef KSH
 		/* set $_ to program's full path */
 		/* setstr() can't fail here */
-		setstr(typeset("_", LOCAL|EXPORT, 0, INTEGER, 0), tp->val.s,
-		       KSH_RETURN_ERROR);
+		setstr(typeset("_", LOCAL|EXPORT, 0, INTEGER, 0),
+		    tp->val.s, KSH_RETURN_ERROR);
 #endif /* KSH */
 
 		if (flags&XEXEC) {
@@ -705,15 +620,15 @@ comexec(struct op *t, struct tbl *volatile tp, char **ap, int volatile flags)
 		texec.left = t;	/* for tprint */
 		texec.str = tp->val.s;
 		texec.args = ap;
-		rv = exchild(&texec, flags, -1);
+		rv = exchild(&texec, flags, xerrok, -1);
 		break;
 	}
-  Leave:
+ Leave:
 	if (flags & XEXEC) {
 		exstat = rv;
 		unwind(LLEAVE);
 	}
-	return rv;
+	return (rv);
 }
 
 static void
@@ -859,7 +774,7 @@ define(const char *name, struct op *t)
 	tp->val.t = tcopy(t->left, tp->areap);
 	tp->flag |= (ISSET|ALLOC);
 
-	return 0;
+	return (0);
 }
 
 /*
@@ -883,9 +798,7 @@ builtin(const char *name, int (*func) (int, char **, int), int flags)
  * either function, hashed command, or built-in (in that order)
  */
 struct tbl *
-findcom(name, flags)
-	const char *name;
-	int	flags;		/* FC_* */
+findcom(const char *name, int flags)
 {
 	static struct tbl temp;
 	struct tbl *tp = NULL, *tbi;
@@ -907,7 +820,7 @@ findcom(name, flags)
 		tp = findfunc(name, FALSE);
 		if (tp && !(tp->flag & ISSET)) {
 			if ((fpath = str_val(global("FPATH"))) == null) {
-				tp->u.fpath = (char *) 0;
+				tp->u.fpath = NULL;
 				tp->u2.errno_ = 0;
 			} else
 				tp->u.fpath = search(name, fpath, R_OK,
@@ -927,10 +840,9 @@ findcom(name, flags)
 	if (!tp && (flags & FC_UNREGBI) && tbi)
 		tp = tbi;
 
-  Search:
-	if ((!tp || (tp->type == CTALIAS && !(tp->flag&ISSET)))
-	    && (flags & FC_PATH))
-	{
+ Search:
+	if ((!tp || (tp->type == CTALIAS && !(tp->flag&ISSET))) &&
+	    (flags & FC_PATH)) {
 		if (!tp) {
 				tp = &temp;
 				tp->type = CEXEC;
@@ -957,7 +869,7 @@ findcom(name, flags)
 			tp->u.fpath = npath;
 		}
 	}
-	return tp;
+	return (tp);
 }
 
 /* Check if path is something we want to find.  Returns -1 for failure. */
@@ -985,7 +897,7 @@ search_access(path, mode, errnop)
 	}
 	if (err && errnop && !*errnop)
 		*errnop = err;
-	return ret;
+	return (ret);
 }
 
 /*
@@ -1007,8 +919,8 @@ search(name, path, mode, errnop)
 		*errnop = 0;
 	if (ksh_strchr_dirsep(name)) {
 		if (search_access(name, mode, errnop) == 0)
-			return (char *) name;
-		return NULL;
+			return (char *)(name);
+		return (NULL);
 	}
 
 	namelen = strlen(name) + 1;
@@ -1028,13 +940,13 @@ search(name, path, mode, errnop)
 		sp = p;
 		XcheckN(xs, xp, namelen);
 		memcpy(xp, name, namelen);
- 		if (search_access(Xstring(xs, xp), mode, errnop) == 0)
+		if (search_access(Xstring(xs, xp), mode, errnop) == 0)
 			return Xclose(xs, xp + namelen);
 		if (*sp++ == '\0')
 			sp = NULL;
 	}
 	Xfree(xs, xp);
-	return NULL;
+	return (NULL);
 }
 
 static int
@@ -1058,8 +970,8 @@ call_builtin(struct tbl *tp, char **wp)
 	shf_flush(shl_stdout);
 	shl_stdout_ok = 0;
 	builtin_flag = 0;
-	builtin_argv0 = (char *) 0;
-	return rv;
+	builtin_argv0 = NULL;
+	return (rv);
 }
 
 /*
@@ -1089,37 +1001,36 @@ iosetup(struct ioword *iop, struct tbl UNUSED(*tp))
 			snptreef((char *) 0, 32, "%R", &iotmp));
 
 	switch (iotype) {
-	  case IOREAD:
+	case IOREAD:
 		flags = O_RDONLY;
 		break;
 
-	  case IOCAT:
+	case IOCAT:
 		flags = O_WRONLY | O_APPEND | O_CREAT;
 		break;
 
-	  case IOWRITE:
+	case IOWRITE:
 		flags = O_WRONLY | O_CREAT | O_TRUNC;
 		/* The stat() is here to allow redirections to
 		 * things like /dev/null without error.
 		 */
-		if (Flag(FNOCLOBBER) && !(iop->flag & IOCLOB)
-		    && (stat(cp, &statb) < 0 || S_ISREG(statb.st_mode)))
+		if (Flag(FNOCLOBBER) && !(iop->flag & IOCLOB) &&
+		    (stat(cp, &statb) < 0 || S_ISREG(statb.st_mode)))
 			flags |= O_EXCL;
 		break;
 
-	  case IORDWR:
+	case IORDWR:
 		flags = O_RDWR | O_CREAT;
 		break;
 
-	  case IOHERE:
+	case IOHERE:
 		do_open = 0;
 		/* herein() returns -2 if error has been printed */
 		u = herein(iop->heredoc, iop->flag & IOEVAL);
 		/* cp may have wrong name */
 		break;
 
-	  case IODUP:
-	  {
+	case IODUP: {
 		const char *emsg;
 
 		do_open = 0;
@@ -1127,13 +1038,14 @@ iosetup(struct ioword *iop, struct tbl UNUSED(*tp))
 			u = 1009;	 /* prevent error return below */
 			do_close = 1;
 		} else if ((u = check_fd(cp,
-				X_OK | ((iop->flag & IORDUP) ? R_OK : W_OK),
-				&emsg)) < 0)
-		{
+		    X_OK | ((iop->flag & IORDUP) ? R_OK : W_OK),
+		    &emsg)) < 0) {
 			warningf(TRUE, "%s: %s",
-				snptreef((char *) 0, 32, "%R", &iotmp), emsg);
-			return -1;
+			    snptreef(NULL, 32, "%R", &iotmp), emsg);
+			return (-1);
 		}
+		if (u == iop->unit)
+			return (0);		/* "dup from" == "dup to" */
 		break;
 	  }
 	}
@@ -1142,33 +1054,44 @@ iosetup(struct ioword *iop, struct tbl UNUSED(*tp))
 	}
 	if (u < 0) {
 		/* herein() may already have printed message */
-		if (u == -1)
+		if (u == -1) {
+			u = errno;
 			warningf(TRUE, "cannot %s %s: %s",
-			       iotype == IODUP ? "dup"
-				: (iotype == IOREAD || iotype == IOHERE) ?
-				    "open" : "create", cp, strerror(errno));
-		return -1;
+			    iotype == IODUP ? "dup" :
+			    (iotype == IOREAD || iotype == IOHERE) ?
+			    "open" : "create", cp, strerror(u));
+		}
+		return (-1);
 	}
 	/* Do not save if it has already been redirected (i.e. "cat >x >y"). */
-	if (e->savefd[iop->unit] == 0)
-		/* posh_builtin_exec() assumes e->savefd[fd] set for any redirections.
-		 * Ask savefd() not to close iop->unit - allows error messages
-		 * to be seen if iop->unit is 2; also means we can't lose
-		 * the fd (eg, both dup2 below and dup2 in restfd() failing).
-		 */
-		e->savefd[iop->unit] = savefd(iop->unit, 1);
+	if (e->savefd[iop->unit] == 0) {
+		/* If these are the same, it means unit was previously closed */
+		if (u == iop->unit)
+			e->savefd[iop->unit] = -1;
+		else
+			/* c_exec() assumes e->savefd[fd] set for any
+			 * redirections. Ask savefd() not to close iop->unit;
+			 * this allows error messages to be seen if iop->unit
+			 * is 2; also means we can't lose the fd (eg, both
+			 * dup2 below and dup2 in restfd() failing).
+			 */
+			e->savefd[iop->unit] = savefd(iop->unit, 1);
+	}
 
 	if (do_close)
 		close(iop->unit);
 	else if (u != iop->unit) {
 		if (ksh_dup2(u, iop->unit, TRUE) < 0) {
+			int ev;
+
+			ev = errno;
 			warningf(TRUE,
-				"could not finish (dup) redirection %s: %s",
-				snptreef((char *) 0, 32, "%R", &iotmp),
-				strerror(errno));
+			    "could not finish (dup) redirection %s: %s",
+			    snptreef(NULL, 32, "%R", &iotmp),
+			    strerror(ev));
 			if (iotype != IODUP)
 				close(u);
-			return -1;
+			return (-1);
 		}
 		if (iotype != IODUP)
 			close(u);
@@ -1186,7 +1109,7 @@ iosetup(struct ioword *iop, struct tbl UNUSED(*tp))
 	}
 	if (u == 2) /* Clear any write errors */
 		shf_reopen(2, SHF_WR, shl_out);
-	return 0;
+	return (0);
 }
 
 /*
@@ -1203,9 +1126,9 @@ herein(const char *content, int sub)
 	int i;
 
 	/* ksh -c 'cat << EOF' can cause this... */
-	if (content == (char *) 0) {
+	if (content == NULL) {
 		warningf(TRUE, "here document missing");
-		return -2; /* special to iosetup(): don't print error */
+		return (-2); /* special to iosetup(): don't print error */
 	}
 
 	/* Create temp file to hold content (done before newenv so temp
@@ -1218,7 +1141,7 @@ herein(const char *content, int sub)
 			h->name, strerror(errno));
 		if (shf)
 			shf_close(shf);
-		return -2 /* special to iosetup(): don't print error */;
+		return (-2) /* special to iosetup(): don't print error */;
 	}
 
 	osource = source;
@@ -1229,7 +1152,7 @@ herein(const char *content, int sub)
 		quitenv();
 		shf_close(shf);	/* after quitenv */
 		close(fd);
-		return -2; /* special to iosetup(): don't print error */
+		return (-2); /* special to iosetup(): don't print error */
 	}
 	if (sub) {
 		/* Do substitutions on the content of heredoc */
@@ -1249,8 +1172,8 @@ herein(const char *content, int sub)
 		close(fd);
 		warningf(TRUE, "error writing %s: %s", h->name,
 			strerror(errno));
-		return -2; /* special to iosetup(): don't print error */
+		return (-2); /* special to iosetup(): don't print error */
 	}
 
-	return fd;
+	return (fd);
 }
