@@ -5,6 +5,22 @@
 #include "sh.h"
 #include <ctype.h>
 
+/*
+ * states while lexing word
+ */
+#define SBASE		0	/* outside any lexical constructs */
+#define SWORD           1	/* implicit quoting for substitute() */
+#define SSQUOTE		2	/* inside '' */
+#define SDQUOTE		3	/* inside "" */
+#define SBRACE		4	/* inside ${} */
+#define SCSPAREN	5	/* inside $() */
+#define SBQUOTE		6	/* inside `` */
+#define SASPAREN	7	/* inside $(( )) */
+#define SHEREDELIM	8	/* parsing <<,<<- delimiter */
+#define SHEREDQUOTE	9	/* parsing " in <<,<<- delimiter */
+#define STBRACE		10	/* parsing ${..[#%]..} */
+
+
 /* Structure to keep track of the lexing state and the various pieces of info
  * needed for each particular state. */
 typedef struct lex_state Lex_state;
@@ -81,22 +97,24 @@ static int ignore_backslash_newline;
 
 /* optimized getsc_bn() */
 #define getsc()		(*source->str != '\0' && *source->str != '\\' \
-			 && !backslash_skip ? *source->str++ : getsc_bn())
-/* optimized getsc__() */
-#define	getsc_()	((*source->str != '\0') ? *source->str++ : getsc__())
+			 && !backslash_skip && !(source->flags & SF_FIRST) \
+			 ? *source->str++ : getsc_bn())
+/* optimised getsc__() */
+#define	getsc_()	((*source->str != '\0') && !(source->flags & SF_FIRST) \
+			 ? *source->str++ : getsc__())
 
 #define STATE_BSIZE	32
 
 #define PUSH_STATE(s)	do { \
-			    if (++statep == state_info.end) \
-				statep = push_state_(&state_info, statep); \
-			    state = statep->ls_state = (s); \
+	if (++statep == state_info.end) \
+		statep = push_state_(&state_info, statep); \
+	state = statep->ls_state = (s); \
 			} while (0)
 
 #define POP_STATE()	do { \
-			    if (--statep == state_info.base) \
-				statep = pop_state_(&state_info, statep); \
-			    state = statep->ls_state; \
+	if (--statep == state_info.base) \
+		statep = pop_state_(&state_info, statep); \
+	state = statep->ls_state; \
 			} while (0)
 
 /*
@@ -178,9 +196,9 @@ yylex(int cf)
 						Source *s;
 
 						s = pushs(SREREAD,
-							  source->areap);
-						s->start = s->str
-							= s->u.freeme = tmp;
+						    source->areap);
+						s->start = s->str =
+						    s->u.freeme = tmp;
 						s->next = source;
 						source = s;
 					}
@@ -601,7 +619,7 @@ yylex(int cf)
 		iop->heredoc = NULL;
 		Xfree(ws, wp);	/* free word */
 		yylval.iop = iop;
-		return REDIR;
+		return (REDIR);
 	}
 
 	if (wp == dp && state == SBASE) {
@@ -651,7 +669,7 @@ yylex(int cf)
 	*wp++ = EOS;		/* terminate word */
 	yylval.cp = Xclose(ws, wp);
 	if (state == SWORD)	/* ONEWORD? */
-		return LWORD;
+		return (LWORD);
 	ungetsc(c);		/* unget terminator */
 
 	/* copy word to unprefixed string ident */
@@ -701,7 +719,7 @@ yylex(int cf)
 		}
 	}
 
-	return LWORD;
+	return (LWORD);
 }
 
 static void
@@ -802,7 +820,7 @@ pushs(int type, Area *areap)
 {
 	Source *s;
 
-	s = (Source *) alloc(sizeof(Source), areap);
+	s = alloc(sizeof(Source), areap);
 	s->type = type;
 	s->str = null;
 	s->start = NULL;
@@ -817,7 +835,7 @@ pushs(int type, Area *areap)
 		Xinit(s->xs, dummy, 256, s->areap);
 	} else
 		memset(&s->xs, 0, sizeof(s->xs));
-	return s;
+	return (s);
 }
 
 static int
@@ -826,6 +844,7 @@ getsc__()
 	Source *s = source;
 	int c;
 
+ getsc_again:
 	while ((c = *s->str++) == 0) {
 		s->str = NULL;		/* return 0 for EOF by default */
 		switch (s->type) {
@@ -918,7 +937,18 @@ getsc__()
 			shf_flush(shl_out);
 		}
 	}
-	return c;
+	/* check for UTF-8 byte order mark */
+	if (s->flags & SF_FIRST) {
+		s->flags &= ~SF_FIRST;
+		if (((unsigned char)c == 0xEF) &&
+		    (((const unsigned char *)(s->str))[0] == 0xBB) &&
+		    (((const unsigned char *)(s->str))[1] == 0xBF)) {
+			s->str += 2;
+			/* UTFMODE = 1; */
+			goto getsc_again;
+		}
+	}
+	return (c);
 }
 
 static void
